/* (C) 2016-2019 Hot-World GmbH & Co. KG <rlittwin@repetier.com>
 All Rights Reserved

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 It is allowed to use this code to create or use in products connecting to
 Repetier-Server or to use it for own plugins/modules with Repetier-Server.
 This includes free and commercial products.
 Usage for other products is strictly forbidden.
 */

import {Injectable, ApplicationRef} from '@angular/core';
import {HttpClient} from '@angular/common/http';
import {SubscriptionLike} from 'rxjs';
import {NgRedux} from '@angular-redux/store';
import {RepetierState} from './state/RepetierState';
import {TemperatureDataState} from './state/PrinterState';
import {UserState} from './state/UserState';
import {ConnectionAction} from './actions/ConnectionAction';
import {ServerAction} from './actions/ServerAction';
import {LicenseAction} from './actions/LicenseAction';
import {PrintersAction} from './actions/PrintersAction';
import {PrinterAction} from './actions/PrinterAction';
import {UserAction} from './actions/UserAction';
import {WifiAction} from './actions/WifiAction';
import {ConnectionHandler, ServerEvent} from './ConnectionHandler';
import {VisualAids} from './VisualAids';
import {Language} from './state/Language';
import {Router} from "@angular/router";

@Injectable()
export class RepetierServer {
    private disposed = false;
    server: string = window.location.host;
    private conDialogVisible = false;

    alert = false;
    conDialog = null;
    private firstConnection = true;
    private eventSubscription: SubscriptionLike;
    private connectionEventSubscription: SubscriptionLike;
    private localEventSubscription: SubscriptionLike;

    private perodical5000Timer: any;
    private perodical2000Timer: any;

    constructor(public app: ApplicationRef,
                private http: HttpClient,
                private redux: NgRedux<RepetierState>,
                private connectionAction: ConnectionAction,
                private serverAction: ServerAction,
                private licenseAction: LicenseAction,
                private connectionHandler: ConnectionHandler,
                private printersAction: PrintersAction,
                private printerAction: PrinterAction,
                private userAction: UserAction,
                private wifiAction: WifiAction,
                private visualAid: VisualAids,
                private router: Router) {

        (<any>window).rs = this;
        this.http.get<Language[]>('/languages.json').subscribe((data) => {
            this.serverAction.setLanguages(data);
        });
        this.eventSubscription = this.connectionHandler.serverEvents.subscribe(this.eventListener);
        this.localEventSubscription = this.connectionHandler.localEvents.subscribe(this.localEventListener);
        this.connectionEventSubscription = this.connectionHandler.connectionEvents.subscribe(this.connectionEventListener);
        this.perodical5000Timer = setInterval(this.periodical5000, 5000);
        this.perodical2000Timer = setInterval(this.periodical2000, 2000);
    }

    periodical5000 = () => {
        if (this.redux.getState().connection.connected) {
            this.printersAction.updatePrinters();
        }
    }

    periodical2000 = () => {
        if (this.redux.getState().connection.connected) {
            this.printersAction.updateStates();
        }
    }

    connectionEventListener = (evt: ServerEvent) => {
        switch (evt.event) {
            case 'updateImportant':
                this.updateImportant();
                break;
        }
    }

    localEventListener = (evt: ServerEvent) => {
        switch (evt.event) {
            case 'connected':
                this.firstConnection = false;
                // Read external commands
                this.connectionHandler.send('listExternalCommands').then((data) => {
                    this.serverAction.setAllExternalCommands(data);
                    this.updateExtCommands(data);
                });
                this.licenseAction.update();
                this.wifiAction.updateRouterList();
                this.serverAction.updateFolders();
                if (this.conDialogVisible) {
                    this.visualAid.onConnected();
                }
                this.serverAction.updateDiskUsage();
                this.conDialogVisible = false;
                this.wifiAction.refreshRouterList();
                break;
            case 'disconnected':
                if (!this.firstConnection) {
                    this.conDialogVisible = true;
                    this.visualAid.onDisconnected();
                }
                break;
        }
    }

    eventListener = (evt: ServerEvent) => {
        const data = evt.data;
        switch (evt.event) {
            case 'settingChanged':
                this.serverAction.setSettings(data);
                /*RSUtils.extend(this.settings, data);
                 this._originalSettings = RSUtils.copyObject(this.settings);*/
                break;
            case 'userCredentials': {
                this.userAction.setUserData({
                    loggedIn: true,
                    login: data.login,
                    settings: data.settings,
                    permissions: data.permissions
                });
                this.updateExtCommands(this.redux.getState().server.allExternalCommands);
                localStorage.setItem('session', JSON.stringify({
                    user: data.login,
                    session: this.redux.getState().connection.session
                }));
                this.connectionHandler.fireLocalEvent('userCredentialsStored');
            }
                break;
            case 'foldersChanged':
                this.serverAction.updateFolders();
                break;
            case 'wifiChanged':
                this.wifiAction.setWifiList(data);
                break;
            case 'progressUpdate':
                this.visualAid.updateProgress(data);
                break;


            // printer related events

            case 'temp': {// new temperatures in O,S,T,id,t(ime)
                if (this.redux.getState().printers.stateTime === 0) {
                    break;
                } // wait for first flood of temperatures to maintain order
                const temp: TemperatureDataState = PrintersAction.prepareTemperature(data);
                if (data.id >= 1000) {
                    if (data.id >= 2000) {
                        this.printerAction.addChamberTemperature(evt.slug, data.id - 2000, temp);
                    } else {
                        this.printerAction.addBedTemperature(evt.slug, data.id - 1000, temp);
                    }
                } else if (data.id < this.redux.getState().printers.printers[evt.slug].extruders.length) {
                    this.printerAction.addExtruderTemperature(evt.slug, data.id, temp);
                }
            }
                break;
            case 'jobsChanged': // gcode modes changed
                this.printerAction.gcodeFetchModels(evt.slug);
                break;
            case 'jobFinished':
                break;
            case 'jobStarted':
                break;
            case 'jobKilled':
                break;
            case 'eepromClear':
                this.printerAction.dispatch(evt.printer.slug, PrinterAction.CLEAR_EEPROM, data);
                break;
            case 'eepromData':
                this.printerAction.dispatch(evt.printer.slug, PrinterAction.APPEND_EEPROM, data);
                break;
            case 'recoverChanged':
                this.printerAction.dispatch(evt.printer.slug, PrinterAction.SET_RECOVER, data);
                break;
            case 'move':
                this.printerAction.updateState(evt.printer.slug, {
                        posX: data.x,
                        posY: data.y,
                        posZ: data.z,
                        posE: data.e,
                        posDE: data.de
                    }
                );
                break;
            case 'log':
                this.printerAction.dispatch(evt.printer.slug, PrinterAction.ADD_LOG_LINE, data);
                break;
            case 'printqueueChanged':
                this.printerAction.gcodeFetchPrintqueue(evt.slug);
                break;
            case 'state': {
                const mod: any = this.printersAction.prepareStateData(evt.slug, data);
                this.printerAction.updateState(evt.slug, mod);
            }
                break;
            case 'config':
                this.printerAction.updateConfig(evt.slug);
                break;
            case 'scriptChanged':
                this.printerAction.updateScript(evt.slug, data);
                break;
            case 'firmwareChanged':
                break;
            case 'printerSettingChanged':
                break;
            case 'licenceChanged':
                this.licenseAction.update();
                break;
            case 'modelGroupListChanged':
                this.printerAction.updateGroups(evt.slug);
                break;
            // Image has changed, initiate update
            case 'newRenderImage': {
                const key = data.slug + '_' + data.list + data.id;
                const urls: { [key: string]: string } = {};
                urls[key + 'l'] = '/dyn/render_image?q=' + data.list + '&id=' + data.id + '&slug=' + data.slug + '&t=l&tm=' + new Date().getTime();
                urls[key + 'm'] = '/dyn/render_image?q=' + data.list + '&id=' + data.id + '&slug=' + data.slug + '&t=m&tm=' + new Date().getTime();
                urls[key + 's'] = '/dyn/render_image?q=' + data.list + '&id=' + data.id + '&slug=' + data.slug + '&t=s&tm=' + new Date().getTime();
                this.printerAction.updateImageUrls(evt.slug, urls);
                if (data.list === 'jobs') {
                    this.printerAction.gcodeFetchPrintqueue(evt.slug);
                } else {
                    this.printerAction.gcodeFetchModels(evt.slug);
                }
            }
                break;
            case 'timer300':
                this.serverAction.updateDiskUsage();
                break;
            case 'changeFilamentRequested':
                this.printersAction.setActiveSlug(evt.slug);
                this.router.navigateByUrl("/changefilament");
                break;
        }
    }

    dispose() {
        this.disposed = true;
        this.connectionEventSubscription.unsubscribe();
    }

    private updateExtCommands(all) {
        const a = [];
        const user: UserState = this.redux.getState().user;
        for (const test of all) {
            if (test.permPrint && !user.canPrint) {
                continue;
            }
            if (test.permConfig && !user.canConfigure) {
                return;
            }
            if (test.permAdd && !user.canAddFiles) {
                return;
            }
            if (test.permDel && !user.canDeleteFiles) {
                return;
            }
            a.push(test);
        }
        this.serverAction.setExternalCommands(a);
    }

    // preload modal helper in case we are offline, we need this for message
    private updateImportant() {
        this.licenseAction.update();
        this.serverAction.loadSettings();
        this.serverAction.updateFolders();
    }

}

