/* (C) 2016-2019 Hot-World GmbH & Co. KG <rlittwin@repetier.com>
 All Rights Reserved

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 It is allowed to use this code to create or use in products connecting to
 Repetier-Server or to use it for own plugins/modules with Repetier-Server.
 This includes free and commercial products.
 Usage for other products is strictly forbidden.
 */

import {Injectable} from "@angular/core";
import {NgRedux} from "@angular-redux/store";
import {RepetierState} from "../state/RepetierState";
import {PrinterErrorState} from "../state/PrinterErrorState";
import {VisualAids} from "../VisualAids";

@Injectable()
export class PrinterErrorAction {

    static ADD_ERROR = "PERROR_ADD_ERROR";
    static UPDATE_ERROR = "PERROR_UPDATE_ERROR";
    static MARK_SEEN = "PERROR_MARK_SEEN";


    static idBed(id: number, slug: string) {
        return "bed" + id + ";" + slug;
    }

    static idChamber(id: number, slug: string) {
        return "chamber" + id + ";" + slug;
    }

    static idExtruder(exid: number, slug: string) {
        return "" + exid + ";" + slug;
    }

    static create(_id: string, slug: string): PrinterErrorState {
        const parts = _id.split(";");
        return {
            id: _id,
            slug: slug,
            seen: true,
            errorId: 0,
            isError: false,
            stateString: "",
            device: parts[0] === "bed" ? "<?php _('Heated Bed') ?>" : "<?php _('Extruder') ?> " + parts[0]
        };
    }

    static errorToString(errorId: number): string {
        switch (errorId) {
            case 0:
                return "<?php _('No Error') ?>";
            case 1:
                return "<?php _('Sensor Defect') ?>";
            case 2:
                return "<?php _('Heater/Sensor Decoupled') ?>";
            case 3:
                return "<?php _('Filament Slipping') ?>";
            case 4:
                return "<?php _('Extruder Jammed') ?>";
            case 5:
                return "<?php _('Extruder Jammed/Out of Filament') ?>";
            case 6:
                return "<?php _('Out of Filament') ?>";
            default:
                return "<?php _('Unknown Error') ?>";
        }
    }

    static listErrors(orig: { [key: string]: PrinterErrorState }): PrinterErrorState[] {
        const a: PrinterErrorState[] = [];
        for (const idx in orig) {
            if (!orig.hasOwnProperty(idx)) {
                continue;
            }
            const e = orig[idx];
            if (!e.isError || e.seen) {
                continue;
            }
            a.push(e);
        }
        a.sort(function (aValue: PrinterErrorState, bValue: PrinterErrorState) {
            const o1 = aValue.device.toLowerCase();
            const o2 = bValue.device.toLowerCase();
            if (o1 < o2) {
                return -1;
            }
            if (o1 > o2) {
                return 1;
            }
            if (aValue.errorId < bValue.errorId) {
                return 1;
            }
            if (aValue.errorId > bValue.errorId) {
                return -1;
            }
            return 0;
        });
        return a;
    }

    constructor(private redux: NgRedux<RepetierState>,
                private visualAids: VisualAids) {
    }

    markSeen(id: string) {
        this.redux.dispatch({type: PrinterErrorAction.MARK_SEEN, payload: id});
    }

    markAllSeen(): void {
        let closeMsg = false;
        const orig: { [key: string]: PrinterErrorState } = this.redux.getState().printerErrors;
        for (const idx in orig) {
            if (!orig.hasOwnProperty(idx)) {
                continue;
            }
            if (orig[idx].seen === false) {
                this.markSeen(idx);
                closeMsg = true;
            }
        }
        if (closeMsg) {
            this.visualAids.onPrinterErrorSolved();
        }
    }

    hasUnseenErrors(): boolean {
        const orig: { [key: string]: PrinterErrorState } = this.redux.getState().printerErrors;
        for (const idx in orig) {
            if (!orig.hasOwnProperty(idx)) {
                continue;
            }
            const e = orig[idx];
            if (!e.isError || e.seen) {
                continue;
            }
            return true;
        }
        return false;
    }

    createIfNeeded(id: string, slug: string): PrinterErrorState {
        let err: PrinterErrorState = this.redux.getState().printerErrors[id];
        if (!err) {
            err = PrinterErrorAction.create(id, slug);
            this.redux.dispatch({type: PrinterErrorAction.ADD_ERROR, payload: err});
        }
        return err;
    }

    updateHeaterError(id: string, e: number): boolean {
        const err: PrinterErrorState = this.redux.getState().printerErrors[id];
        if (!err) {
            return false;
        }
        if (e === err.errorId) {
            return false;
        }
        this.redux.dispatch({
            type: PrinterErrorAction.UPDATE_ERROR, payload: {
                id: id, error: {
                    isError: e !== 0,
                    errorId: e,
                    seen: e === 0,
                    stateString: PrinterErrorAction.errorToString(e)
                }
            }
        });
        this.visualAids.onPrinterError();
        return true;
    }
}
