/* (C) 2016-2019 Hot-World GmbH & Co. KG <rlittwin@repetier.com>
 All Rights Reserved

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 It is allowed to use this code to create or use in products connecting to
 Repetier-Server or to use it for own plugins/modules with Repetier-Server.
 This includes free and commercial products.
 Usage for other products is strictly forbidden.
 */

import {Injectable} from '@angular/core';
import {NgRedux} from '@angular-redux/store';
import {ConnectionHandler, ServerEvent} from '../ConnectionHandler';
import {RepetierState} from '../state/RepetierState';
import {PrintersState} from '../state/PrintersState';
import {PrinterErrorAction} from './PrinterErrorAction';
import {
    PrinterState,
    INITIAL_PRINTER_STATE,
    ToolState,
    INITAL_TOOL_STATE,
    TemperatureDataState
} from '../state/PrinterState';
import {PrinterAction} from './PrinterAction';
import {VisualAids} from '../VisualAids';
import {RSUtils} from '../RSUtils';

@Injectable()
export class PrintersAction {

    static SET_ACTIVE = 'PRINTERS_SET_ACTIVE';
    static CREATE_PRINTER = 'PRINTERS_CREATE';
    static DESTROY_PRINTER = 'PRINTERS_DESTROY';
    static SET_AUTOSELECT = 'PRINTERS_SET_AUTOSELECT';
    static SET_STATE_TIME = 'PRINTER_SET_STATE_TIME';

    private updatingState = false;
    private updatingStatus = false;
    private sendMovesCounter = 0;

    static prepareTemperature(temp: any): TemperatureDataState {
        return {
            output: temp.O,
            time: temp.t,
            tempSet: temp.S,
            temp: temp.T
        };
    }

    private static preparePrinterListData(data): any {
        const out: any = {
            slug: data.slug,
            online: data.online,
            active: data.active,
            name: data.name
        };
        if (data.job === 'none') {
            out.printJob = '';
            out.printing = false;
            out.paused = false;
            out.printDone = 0;
        } else {
            out.printJob = data.job;
            out.printing = true;
            out.printDone = data.done;
            out.jobid = data.jobid;
            out.paused = data.paused;
            out.pauseState = data.pauseState;
            out.ofLayer = data.ofLayer;
            out.printedTimeComp = data.printedTimeComp;
            out.printedTime = data.printTime;
            out.startTime = data.start;
        }
        out.idle = !out.printing || out.paused;
        return out;
    }

    private static prepareToolFromState(tool: ToolState, data: any) {
        if (data.history) {
            for (const temp of data.history) {
                const newT: TemperatureDataState = PrintersAction.prepareTemperature(temp);
                while (tool.tempHistory.length > 0 && tool.tempHistory[0].time + 300000 < newT.time) {
                    tool.tempHistory.splice(0, 1);
                }
                tool.tempHistory.push(newT);
            }
        }
        tool.currentTemp = data.tempRead;
        tool.tempSet = data.tempSet;
        tool.output = data.output;
    }

    constructor(private redux: NgRedux<RepetierState>,
                private connectionHandler: ConnectionHandler,
                private visualAid: VisualAids,
                private printerAction: PrinterAction,
                private printerErrorAction: PrinterErrorAction) {
        this.connectionHandler.connectionState.subscribe(connected => {
            if (connected) {
                this.updatingState = false;
                this.updatingStatus = false;
                if (this.sendMovesCounter > 0) {
                    this.connectionHandler.send('sendMoves');
                }
                const pList = this.redux.getState().printers.printers;
                for (const p in pList) {
                    if (pList.hasOwnProperty(p)) {
                        this.printerAction.sendLogLevel(p);
                        this.printerAction.resetImageReference(p);
                    }
                }
                this.updatePrinters();
            }
        });
        this.connectionHandler.localEvents.subscribe(this.localEventHandler);
    }

    private localEventHandler = (e: ServerEvent) => {
        switch (e.event) {
            case 'updateStates':
                this.updateStates();
                break;
        }
    }

    setActiveSlug(slug: string): void {
        this.redux.dispatch({type: PrintersAction.SET_ACTIVE, payload: {slug: slug}});
    }

    setAutoselect(on: boolean): void {
        this.redux.dispatch({type: PrintersAction.SET_AUTOSELECT, payload: on});
    }

    addPrinter(printer: PrintersState): void {
        this.redux.dispatch({type: PrintersAction.CREATE_PRINTER, payload: printer});
    }

    delPrinter(slug: string): void {
        this.redux.dispatch({type: PrintersAction.DESTROY_PRINTER, payload: {slug: slug}});
    }

    public prepareStateData(slug: string, data: any): any {
        const printer: PrinterState = this.redux.getState().printers.printers[slug];
        if (!printer) {
            return {};
        }
        const out: any = {};
        out.fans = RSUtils.copyArray(data.fans);
        out.fans.forEach(x => {
            x.percent = Math.floor(0.5 + x.voltage / 2.55);
        });
        out.activeExtruderId = data.activeExtruder;
        out.debugLevel = data.debugLevel;
        out.firmware = data.firmware;
        out.firmwareURL = data.firmwareURL;
        out.speedMultiply = data.speedMultiply;
        out.flowMultiply = data.flowMultiply;
        out.hasXHome = data.hasXHome;
        out.hasYHome = data.hasYHome;
        out.hasZHome = data.hasZHome;
        out.layer = data.layer;
        out.powerOn = data.powerOn;
        out.sdcardMounted = data.sdcardMounted;
        out.rec = data.rec;
        out.volumetric = data.volumetric;
        out.lightsOn = data.lights;
        out.posX = data.x;
        out.posY = data.y;
        out.posZ = data.z;
        out.extruders = RSUtils.copyArray(printer.extruders);
        out.heatedBeds = RSUtils.copyArray(printer.heatedBeds);
        out.heatedChambers = RSUtils.copyArray(printer.heatedChambers);
        let tempError = false;
        if (data.numExtruder < printer.numExtruder) {
            out.extruders.splice(data.numExtruder, out.extruders.length - data.numExtruder);
        }
        while (data.numExtruder > out.extruders.length) {
            const t: ToolState = RSUtils.copyObject(INITAL_TOOL_STATE);
            t.id = out.extruders.length;
            out.extruders.push(t);
        }
        out.numExtruder = data.numExtruder;
        const numBeds = printer.config.heatedBeds.length;
        const numChambers = printer.config.heatedChambers.length;
        if (data.heatedBeds.length < numBeds) {
            out.heatedBeds.splice(numBeds, data.heatedBeds.length - numBeds);
        }
        while (numBeds > out.heatedBeds.length) {
            const t: ToolState = RSUtils.copyObject(INITAL_TOOL_STATE);
            t.id = out.heatedBeds.length + 1000;
            out.heatedBeds.push(t);
        }
        if (data.heatedChambers.length < numChambers) {
            out.heatedChambers.splice(numChambers, data.heatedChambers.length - numChambers);
        }
        while (numChambers > out.heatedChambers.length) {
            const t: ToolState = RSUtils.copyObject(INITAL_TOOL_STATE);
            t.id = out.heatedChambers.length + 2000;
            out.heatedChambers.push(t);
        }
        for (let i = 0; i < out.numExtruder; i++) {
            PrintersAction.prepareToolFromState(out.extruders[i], data.extruder[i]);
            const bid = PrinterErrorAction.idExtruder(i, slug);
            this.printerErrorAction.createIfNeeded(bid, slug);
            tempError = tempError || this.printerErrorAction.updateHeaterError(bid, out.extruders[i].error);
        }
        if (data.heatedBeds) {
            for (let i = 0; i < numBeds; i++) {
                PrintersAction.prepareToolFromState(out.heatedBeds[i], data.heatedBeds[i]);
                const bid = PrinterErrorAction.idBed(i, slug);
                this.printerErrorAction.createIfNeeded(bid, slug);
                tempError = tempError || this.printerErrorAction.updateHeaterError(bid, data.heatedBeds[i].error);
            }
        }
        if (data.heatedChambers) {
            for (let i = 0; i < numChambers; i++) {
                PrintersAction.prepareToolFromState(out.heatedChambers[i], data.heatedChambers[i]);
                const bid = PrinterErrorAction.idChamber(i, slug);
                this.printerErrorAction.createIfNeeded(bid, slug);
                tempError = tempError || this.printerErrorAction.updateHeaterError(bid, data.heatedChambers[i].error);
            }
        }
        if (tempError) {
            if (this.printerErrorAction.hasUnseenErrors()) {
                this.visualAid.onPrinterError();
            } else {
                this.visualAid.onPrinterErrorSolved();
            }
        }
        return out;
    }

    updatePrinters = () => {
        if (this.updatingStatus) {
            return;
        } // no double query
        this.updatingStatus = true;
        this.connectionHandler.send('listPrinter').then((data) => {
            const exists: { [key: string]: boolean } = {};
            const printers: PrintersState = this.redux.getState().printers;
            for (const p in printers.printers) {
                if (printers.printers.hasOwnProperty(p)) {
                    exists[p] = false;
                }
            }
            for (const p of data) {
                if (!printers.printers[p.slug]) {
                    exists[p.slug] = true;
                    this.addPrinter(Object.assign({}, INITIAL_PRINTER_STATE, PrintersAction.preparePrinterListData(p)));
                    this.printerAction.sendLogLevel(p.slug);
                    this.printerAction.updateData(p.slug);
                } else {
                    exists[p.slug] = true;
                    const newStatus = PrintersAction.preparePrinterListData(p);
                    if (RSUtils.newChanges(printers.printers[p.slug], newStatus)) {
                        this.printerAction.updateStatus(p.slug, newStatus);
                    }
                }
            }
            for (const key in exists) {
                if (!exists.hasOwnProperty(key)) {
                    continue;
                }
                if (exists[key] === false) {
                    this.delPrinter(key);
                }
            }
            this.updatingStatus = false;
        });
    }

    updateStates = () => {
        if (this.updatingState) {
            return;
        }
        this.updatingState = true;
        this.connectionHandler.send('stateList', {includeHistory: this.redux.getState().printers.stateTime === 0}).then((data) => {
            const stateTime = new Date().getTime(); // Time of last state change
            let update = false;
            for (const key in data) {
                if (!data.hasOwnProperty(key)) {
                    continue;
                }
                if (this.redux.getState().printers.printers[key]) {
                    const newState = this.prepareStateData(key, data[key]);
                    if (RSUtils.newChanges(this.redux.getState().printers.printers[key], newState)) {
                        this.printerAction.updateState(key, newState);
                        update = true;
                    }
                }
            }
            this.updatingState = false;
            if (update) {
                this.redux.dispatch({type: PrintersAction.SET_STATE_TIME, payload: stateTime});
            }
        });
    }

    sendMoves() {
        if (this.sendMovesCounter === 0) {
            this.connectionHandler.send('sendMoves');
        }
        this.sendMovesCounter++;
    }

    hideMoves() {
        if (this.sendMovesCounter === 1) {
            this.connectionHandler.send('hideMoves');
        }
        if (this.sendMovesCounter > 0) {
            this.sendMovesCounter--;
        }
    }
}
