/* (C) 2016-2019 Hot-World GmbH & Co. KG <rlittwin@repetier.com>
 All Rights Reserved

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 It is allowed to use this code to create or use in products connecting to
 Repetier-Server or to use it for own plugins/modules with Repetier-Server.
 This includes free and commercial products.
 Usage for other products is strictly forbidden.
 */

import {
    Component,
    Input,
    ElementRef,
    ChangeDetectionStrategy,
    ChangeDetectorRef,
    Inject, OnDestroy, HostBinding
} from '@angular/core';
import {RepetierServer, TemperatureData, PrintersState, PrinterState} from 'binding';
import {select} from '@angular-redux/store';
import {Observable, SubscriptionLike} from 'rxjs';

@Component({
    selector: 'rs-temp-graph',
    template: `
      <svg *ngIf="ok" [attr.width]="width" [attr.height]="height" preserveAspectRatio="xMidYMid meet">
        <path [attr.d]="grid" class="grid"/>
        <path [attr.d]="axis" class="axis"/>
        <path [attr.d]="ticks" class="ticks"/>
        <path [attr.d]="output" class="output"/>
        <path [attr.d]="setTemp" class="tempSet"/>
        <path [attr.d]="curTemp" class="temp"/>
        <text *ngFor="let t of tempLegend" [attr.x]="t.x" [attr.y]="t.y" class="tempAxisText">{{t.t}}</text>
        <text *ngFor="let t of timeLegend" [attr.x]="t.x" [attr.y]="t.y" class="timeAxisText">{{t.t}}</text>
        <text *ngFor="let t of outputLegend" [attr.x]="t.x" [attr.y]="t.y" class="outputAxisText">{{t.t}}</text>
      </svg>
    `,
    changeDetection: ChangeDetectionStrategy.OnPush
})
export class RSTempGraphComponent implements OnDestroy {
    @HostBinding('class.graph') clsGraph = true;
    @Input() extruder: number = undefined;
    ok = false;
    width = 1;
    height = 1;
    minTemp = 0;
    maxTemp = 300;
    maxTime: number;
    minTime: number;
    grid: string;
    axis: string;
    ticks: string;
    setTemp: string;
    curTemp: string;
    output: string;
    private left: number;
    private right: number;
    private top: number;
    private bottom: number;
    private lastTime = 0;
    tempLegend: any = [];
    outputLegend: any = [];
    timeLegend: any = [];
    @Input() private tickLength = 5;
    @select('printers') printers$: Observable<PrintersState>;
    active: PrinterState;
    private subscription: SubscriptionLike;

    graph: TemperatureData[]; // output / temp / tempSet / time

    static twoDigits(a) {
        a = a.toFixed(0);
        if (a.length === 1) {
            return '0' + a;
        }
        return a;
    }

    constructor(public server: RepetierServer,
                @Inject(ElementRef) public elementRef: ElementRef,
                private changeRef: ChangeDetectorRef) {
        this.subscription = this.printers$.subscribe((p: PrintersState) => {
            if (p.active) {
                this.active = p.printers[p.active];
            }
            this.updateGraph();
            this.changeRef.markForCheck();
        });
    }

    ngOnDestroy(): void {
        this.subscription.unsubscribe();
    }

    private timeToX(t: number): number {
        return this.left + (t - this.minTime) * (this.right - this.left) / (this.maxTime - this.minTime);
    }

    private tempToY(t: number): number {
        return this.bottom - (t - this.minTemp) * (this.bottom - this.top) / (this.maxTemp - this.minTemp);
    }

    private outputToY(t: number): number {
        return this.bottom - t * (this.bottom - this.top) / 100;
    }


    updateGraph() {
        if (this.extruder === undefined) {
            this.ok = false;
            return;
        }
        this.graph = [];
        if (this.extruder >= 2000) { // heated chamber
            if (this.active && this.active.heatedChambers && this.active.heatedChambers[this.extruder - 2000].tempHistory) {
                this.graph = this.active.heatedChambers[this.extruder - 2000].tempHistory;
            }
        } else if (this.extruder >= 1000) { // heated bed
            if (this.active && this.active.heatedBeds && this.active.heatedBeds[this.extruder - 1000].tempHistory) {
                this.graph = this.active.heatedBeds[this.extruder - 1000].tempHistory;
            }
        } else {
            if (this.active && this.active.extruders && this.active.extruders.length > this.extruder && this.active.extruders[this.extruder].tempHistory) {
                this.graph = this.active.extruders[this.extruder].tempHistory;
            }
        }
        if (this.graph.length === 0) {
            this.ok = false;
            return;
        }
        if (this.lastTime === this.graph[this.graph.length - 1].time) {
            return;
        } // no new data
        this.minTemp = 99999;
        this.maxTemp = -99999;
        this.maxTime = this.graph[this.graph.length - 1].time;
        this.minTime = this.maxTime - 1000 * 5 * 60; // 5 minutes display
        for (const idx of Object.keys(this.graph)) {
            const act = this.graph[idx];
            if (act.temp < this.minTemp) {
                this.minTemp = act.temp;
            }
            if (act.temp > this.maxTemp) {
                this.maxTemp = act.temp;
            }
            if (act.tempSet < this.minTemp) {
                this.minTemp = act.tempSet;
            }
            if (act.tempSet > this.maxTemp) {
                this.maxTemp = act.tempSet;
            }
        }
        if (this.minTemp < 0) {
            this.minTemp = 0;
        } // minimum is 0
        this.minTemp = Math.max(0, Math.round((this.minTemp - 12) / 10) * 10) - 5;
        this.maxTemp = Math.max(this.minTemp + 20, Math.round((this.maxTemp + 12) / 10) * 10);
        const w: number = this.elementRef.nativeElement.offsetWidth;
        const h: number = this.elementRef.nativeElement.parentElement.offsetHeight;
        this.width = w;
        this.height = h;
        const showOutLegend = w > 200;
        this.left = 50;
        this.right = showOutLegend ? w - 45 : w - this.tickLength;
        this.bottom = h - 25;
        this.top = 10;
        const ticks = [];
        this.axis = ['M', this.left, this.top, 'L', this.left, this.bottom, 'L', this.right, this.bottom, 'L', this.right, this.top,
            'L', this.left, this.top
        ].join(' ');
        const grid = [];
        let actTime = Math.round(this.minTime / 60000) * 60000;
        this.timeLegend = [];
        do {
            if (actTime > this.minTime && actTime < this.maxTime) {
                const xt = this.timeToX(actTime);
                grid.push('M');
                grid.push(xt);
                grid.push(this.bottom);
                grid.push('L');
                grid.push(xt);
                grid.push(this.top);
                ticks.push('M');
                ticks.push(xt);
                ticks.push(this.bottom);
                ticks.push('L');
                ticks.push(xt);
                ticks.push(this.bottom + this.tickLength);
                const d: Date = new Date(actTime);
                this.timeLegend.push({
                    x: xt,
                    y: this.height - 5,
                    t: '' + d.getHours() + ':' + RSTempGraphComponent.twoDigits(d.getMinutes())
                });
            }
            actTime += 60000;
        } while (actTime < this.maxTime);
        let actTemp = this.minTemp + 5;
        do {
            const yt = this.tempToY(actTemp);
            if (yt < this.bottom && yt > this.top) {
                grid.push('M');
                grid.push(this.left);
                grid.push(yt);
                grid.push('L');
                grid.push(this.right);
                grid.push(yt);
                ticks.push('M');
                ticks.push(this.left);
                ticks.push(yt);
                ticks.push('L');
                ticks.push(this.left - this.tickLength);
                ticks.push(yt);
            }
            actTemp += 10;
        } while (actTemp < this.maxTemp);

        // Output ticks
        this.outputLegend = [];
        for (let i = 0; i <= 100; i += 25) {
            const yt = this.outputToY(i);
            ticks.push('M');
            ticks.push(this.right);
            ticks.push(yt);
            ticks.push('L');
            ticks.push(this.right + this.tickLength);
            ticks.push(yt);
            if (showOutLegend) {
                this.outputLegend.push({x: this.right + 5 + this.tickLength, y: yt, t: '' + i + '%'});
            }
        }

        this.grid = grid.join(' ');
        this.ticks = ticks.join(' ');
        const temp = [];
        const tempSet = [];
        const out = [];
        let lastX = 0;
        let xp;
        for (const idx of Object.keys(this.graph)) {
            const t: TemperatureData = this.graph[idx];
            xp = this.timeToX(t.time);
            if (lastX > xp) {
                continue;
            }
            lastX = xp;
            if (xp >= this.left) {
                if (temp.length === 0) {
                    temp.push('M');
                    tempSet.push('M');
                    out.push('M');
                } else {
                    temp.push('L');
                    tempSet.push('L');
                    out.push('L');
                }
                temp.push(xp);
                tempSet.push(xp);
                out.push(xp);
                temp.push(this.tempToY(t.temp));
                tempSet.push(this.tempToY(t.tempSet));
                out.push(this.outputToY(t.output));
            }
        }
        out.push('L');
        out.push(xp);
        out.push(this.bottom);
        out.push('L');
        out.push(out[1]);
        out.push(this.bottom);
        out.push('L');
        out.push(out[1]);
        out.push(out[2]);
        this.curTemp = temp.join(' ');
        this.setTemp = tempSet.join(' ');
        this.output = out.join(' ');
        // write legends
        this.tempLegend = [];
        const dy = Math.abs(this.tempToY(90) - this.tempToY(100));
        const every = Math.floor(15 / dy) + 1;
        actTemp = this.minTemp + 5;
        while (actTemp <= this.maxTemp) {
            this.tempLegend.push(
                {
                    x: this.left - this.tickLength - 5,
                    y: this.tempToY(actTemp),
                    t: '' + actTemp + '°C'
                }
            );
            actTemp += every * 10;
        }

        this.ok = true;
        this.lastTime = this.graph[this.graph.length - 1].time;
    }
}
